/* PHAVerLite: PHAVer + PPLite.
   Copyright (C) 2018 Goran Frehse <goranf@gmail.com>
   Copyright (C) 2019 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PHAVerLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <phaverlite-config.h>

#include "symb_states.hh"
#include "stopwatch.hh"

symb_state_maplist::symb_state_maplist(const symb_states_type& states,
                                       symb_state_plist& newstates) {
  initialize(states, newstates, false);
}

void
symb_state_maplist::clear() {
  state_list.clear();
  iter_map.clear();
}

void
symb_state_maplist::initialize(const symb_states_type& states,
                               symb_state_plist& newstates,
                               bool use_convex_hull) {
  clear();
  for (const auto& p : states) {
    auto iter = add(p.first, p.second, use_convex_hull);
    if (iter != end() && newstates.find(iter) == newstates.end())
      newstates.push_back(iter);
  }
}

symb_state_maplist::iterator
symb_state_maplist::erase(iterator it) {
  // first remove it from iter_map
  loc_ref loc = it->loc;
  auto m_iter = iter_map.find(loc);
  if (m_iter != iter_map.end()) {
    m_iter->second.remove(it);
    if (m_iter->second.empty())
      iter_map.erase(loc);
  }
  return state_list.erase(it);
}

symb_state_maplist::iterator
symb_state_maplist::add(loc_ref loc, clock_val_set new_cvs,
                        bool use_convex_hull) {
  // Check if iter_map already contains loc.
  auto m_iter = iter_map.find(loc);
  if (m_iter == iter_map.end()) {
    // Not found: add it to front of list.
    if (use_convex_hull)
      new_cvs.convex_hull_assign();
    if (MEMORY_MODE >= 4)
      new_cvs.minimize_memory();
    state_list.emplace_front(loc, std::move(new_cvs));
    auto it = state_list.begin();
    iter_map[loc].push_front(it);
    return it;
  }

  // Found loc: get its iter list.
  auto& m_list = m_iter->second;
  // Check for subsumption.
  for (auto it : m_list) {
    if (it->cvs.contains_return_others(new_cvs))
      return end();
  }
  // Not subsumed: add it.
  if (use_convex_hull) {
    assert(m_list.size() == 1);
    auto it = m_list.front();
    auto& cvs = it->cvs;
    cvs.convex_hull_assign(new_cvs);
    if (MEMORY_MODE >= 4)
      cvs.minimize_memory();
    return it;
  }
  // Not using convex hull
  if (MEMORY_MODE >= 4)
    new_cvs.minimize_memory();
  state_list.emplace_front(loc, std::move(new_cvs));
  auto it = state_list.begin();
  m_list.push_front(it);
  return it;
}

bool
symb_state_maplist::is_empty(loc_ref loc) const {
  return iter_map.find(loc) == iter_map.end();
}

bool
symb_state_maplist::is_disjoint_from(const symb_states_type& states) const {
  for (const auto& p : states) {
    loc_ref loc = p.first;
    const auto& cvs = p.second;
    auto m_iter = iter_map.find(loc);
    if (m_iter != iter_map.end()) {
      auto& m_list = m_iter->second;
      for (auto it : m_list) {
        if (!cvs.is_disjoint_from(it->cvs))
          return false;
      }
    }
  }
  return true;
}

symb_state_plist
symb_state_maplist::get_all_states() {
  symb_state_plist res;
  for (auto i = begin(); i != end(); ++i)
    res.push_back(i);
  return res;
}

void
symb_state_maplist::print() {
  for (auto p : iter_map) {
    std::cout << "Location " << p.first << ":" << std::endl;
    for (auto it : p.second)
      it->print();
  }
}

size_t
symb_state_maplist::cvs_size() const {
  size_t sz = 0;
  for (auto p : iter_map)
    for (auto it : p.second)
      sz += it->cvs.size();
  return sz;
}

clock_val_set
symb_state_maplist::get_clock_val_set(loc_ref loc) const {
  if (state_list.empty())
    return clock_val_set(0, Spec_Elem::EMPTY);
  auto m_iter = iter_map.find(loc);
  if (m_iter == iter_map.end()) {
    auto sd = state_list.front().cvs.dim;
    return clock_val_set(sd, Spec_Elem::EMPTY);
  }
  // *this is not empty, and iter_map contains loc
  const auto& m_list = m_iter->second;
  assert(!m_list.empty());
  // Copy first.
  auto it = m_list.begin();
  clock_val_set res = (*it)->cvs;
  // Merge others (starting from second).
  for (++it; it != m_list.end(); ++it)
    res.union_assign((*it)->cvs);
  return res;
}

symb_states_type
symb_state_maplist::transfer_symb_states() {
  symb_states_type states;
  for (auto p : iter_map) {
    loc_ref loc = p.first;
    states[loc] = get_clock_val_set(loc);
  }
  clear();
  return states;
}

void
symb_state_maplist::split(loc_ref oldloc,
                          const clock_val_set& oldinv,
                          loc_ref newloc,
                          const clock_val_set& newinv,
                          symb_state_plist& check_states,
                          symb_state_plist& new_states,
                          bool use_convex_hull) {
  clock_val_set cvs;

  // ENEA: FIXME: why iterating the list without using the map?
  for (auto it_old = begin(); it_old != end(); ) {
    if (it_old->loc != oldloc) {
      ++it_old;
      continue;
    }
    // add intersection with newinv to maplist and get iterator itnew
    cvs.intersection_assign_from(it_old->cvs, newinv);
    auto it_new = cvs.is_empty()
      ? end()
      : add(newloc, std::move(cvs), use_convex_hull);
    if (it_new != end()) {
      // intersect with invariant of oldloc
      it_old->cvs.intersection_assign(oldinv);
      // if it_old is in newstates/checkstates, add it_new
      if (new_states.contains(it_old) && !new_states.contains(it_new))
        new_states.insert(new_states.find(it_old), it_new);
      if (check_states.contains(it_old) && !check_states.contains(it_new))
        check_states.insert(check_states.find(it_old), it_new);
    }

    // erase it_old if it's empty now
    if (!it_old->cvs.is_empty())
      ++it_old;
    else {
      // erase it.
      new_states.remove(it_old);
      check_states.remove(it_old);
      it_old = erase(it_old);
    }
  }
}
