/* PHAVerLite: PHAVer + PPLite.
   Copyright (C) 2018 Goran Frehse <goranf@gmail.com>
   Copyright (C) 2019 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PHAVerLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GUARD_symb_states_hh
#define GUARD_symb_states_hh

#include "clock_val_set.hh"
#include "symb_states_type.hh"

#include <algorithm>
#include <iostream>
#include <list>
#include <map>
#include <set>
#include <vector>
#include <stdexcept>

struct symb_state {
  loc_ref loc;
  clock_val_set cvs;

  symb_state(loc_ref loc_, clock_val_set cvs_)
    : loc(loc_), cvs(std::move(cvs_)) {}

  void print() {
    std::cout << "[" << loc << "]:" << std::endl;
    cvs.print();
  }
};

struct symb_state_plist : public std::list<std::list<symb_state>::iterator> {
  using value_type = std::list<symb_state>::iterator;
  using Base = std::list<value_type>;
  using Iter = Base::iterator;

  bool contains(value_type key) const {
    return std::find(begin(), end(), key) != end();
  }

  Iter find(value_type key) {
    return std::find(begin(), end(), key);
  }

  void print() const {
    for (auto sit : *this)
      sit->print();
  }
}; // symb_state_plist

struct symb_state_maplist {
  // Note the maplist is designed to be strictly increasing.
  // The entries must not be changed, otherwise the pointers
  // in any associated symb_state_plist are invalidated.
  using List = std::list<symb_state>;
  using Map = std::map<loc_ref, symb_state_plist>;

  List state_list;
  Map iter_map;

  symb_state_maplist() = default;
  symb_state_maplist(const symb_states_type& states,
                     symb_state_plist& newstates);

  void clear();
  void initialize(const symb_states_type& states,
                  symb_state_plist& newstates,
                  bool use_convex_hull);

  using iterator = List::iterator;
  using const_iterator = List::const_iterator;

  iterator begin() { return state_list.begin(); }
  iterator end() { return state_list.end(); }
  const_iterator begin() const { return state_list.begin(); }
  const_iterator end() const { return state_list.end(); }

  iterator erase(iterator it);
  iterator add(loc_ref loc, clock_val_set cvs, bool use_convex_hull);

  void split(loc_ref oldloc, const clock_val_set& oldinv,
             loc_ref newloc, const clock_val_set& newinv,
             symb_state_plist& check_states,
             symb_state_plist& new_states,
             bool convex);

  void print();

  bool is_empty(loc_ref loc) const;
  bool is_disjoint_from(const symb_states_type& states) const;
  size_t cvs_size() const;
  size_t loc_size() const { return iter_map.size(); }
  clock_val_set get_clock_val_set(loc_ref loc) const;

  symb_state_plist get_all_states();
  symb_states_type transfer_symb_states();
}; // symb_state_maplist

NOTHROW_MOVES(symb_state);
NOTHROW_DEFAULT_AND_MOVES(symb_state_plist);
NOTHROW_MOVES(symb_state_maplist);

void
insert_maplist_iter(const symb_state_maplist& states,
                    symb_state_plist& new_states,
                    symb_state_maplist::iterator new_iter);

#endif // GUARD_symb_states_hh
